#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <fcntl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <termio.h>

#include <string>

using namespace std;

static const char *ownnumber = "27351010";
static const char *ISDNDevice = "/dev/ttyI0";
static const int NumRingsToWait = 4;

static char code[] = "\n"
":INIT\n"
"p /usr/src/rsynth-2.0/say -r 8000 -a -o -\n"
"P -S 0.7 -c 0 -x 2800 'tss puhuu joelin puhelinvastaaja.'\n"
"P -S 0.7 -c 0 -x 2800 'paina 5i jos haluat jtt viestin,'\n"
/*
"P -S 0.7 -c 0 -x 2800 'tai 2i jos haluat tiet miss joel on,'\n"
"P -S 0.7 -c 0 -x 2800 'tai 3 jos haluat kuulla tmn viestin uudestaan.'\n"
"P -S 0.7 -c 0 -x 2800 'tmn viestin ihan oikeasti lausui tietokoneohjelma.'\n"
"#\n"
"\\5 G MSG\n"
"\\2 G WHERE\n"
"\\3 G INIT\n"
"\\9 G QUIT\n"
"\\4 G VISA\n"
"\\0 G QUIT\n"
":QUIT\n"
"P -S 0.9 -c 2 -x 600 ' suljessepuhelin'\n"
"Q\n"
":WHERE\n"
"P -S 0.9 -c 2 -x 1100 'ei tavoitettavissa puhelimitse. jos haluat jtt hnelle viestin, paina 5i.'\n"
"#\n"
"\\5 G MSG\n"
"\\3 G INIT\n"
"\\9 G QUIT\n"
"\\0 G QUIT\n"
":MSG\n"
"P -S 0.9 -c 3 -x 2100 'viestill ei ole pituusrajaa.'\n"
"P -S 0.9 -c 3 -x 1800 'paina 0 tai sulje puhelin kun lopetat.'\n"
"I\n"*/
"Q\n"
;

static int fd;

struct termio back;

#define DLE 0x10
#define ETX 0x13
#define DC4 0x14

static void Open(const char *s)
{
	struct termio term;
	
	fd = open(s, O_RDWR);
	
	ioctl(0, TCGETA, &back);
	
	term=back;
    term.c_lflag &= ~(ICANON|ISIG|ECHO);	/* enable ^C */
    term.c_iflag &= ~(IXON);				/* enable ^Q */
	term.c_cc[VMIN] = 1;
	ioctl(0, TCSETA, &term);

	fcntl(0,  F_SETFL,  fcntl(0, F_GETFL) | O_NONBLOCK);
	fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
}

static void Close(void)
{
	close(fd);
	ioctl(0, TCSETA, &back);
}

static void Send(const void *p, unsigned size)
{ 	
	const char *q = (const char *)p;
	fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) & ~O_NONBLOCK);
	while(size)
	{
		int n = write(fd, q, size);
		if(n < 0)
		{
			printf("Ouch!\n");
			perror("term");
			return;
		}
		size -= n;
		q += n;
	}
	fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
}

static void Send(int i)
{
	char c = i;
	Send(&c, 1);
}

static void Send(const string &s)
{
	printf("Sending '%s'\n", s.c_str());
	
	Send(s.c_str(), s.size());
}

static void InitTerm(const char *mynumber)
{
	Send(
		"ATZ\n"
		"AT+FCLASS=8\n"
		"AT+VIP\n"
		"ATE1 S0=0 H0 Q0 V1\n");
		
	Send("AT&E"); Send(mynumber);
	Send("\nAT&L"); Send(mynumber);
	Send("\n");
}

static void Term(void)
{
/*
*** Bytes:
	   ?? = 0x03
      DLE = 0x10
      ETX = 0x13
      DC4 = 0x14
      
     send:
      DLE-DLE = output DLE                      ^P^P
      DLE-ETX = end of data, vonline |= 4       ^P^S
      DLE-DC4 = abort RX,    vonline &= ~1      ^P^T
      DLE-0x03= abort TX                        ^P^C
      DLE-'s' and
      DLE-'q' = silence (?)
      
     received:
      DLE-num = received DTMF code (0..15) ?
      
*** Results:
      0=OK
      1=CONNECT
      2=RING
      3=NO CARRIER
      4=ERROR
      5=CONNECT 64000
      6=NO DIALTONE
      7=BUSY
      8=NO ANSWER
      9=RINGING
      10=NO MSN/EAZ
      11=VCON
      12=RUNG
      
*** Available commands:
      ATA       - answer.
      ATD ...   - dial
                  accepted chars:
                     0-9
                     #*
                     S         (at start only)
      ATE x     - echo x[0=off,1=on]
      ATH x     - hook x[0=on,1=off]
      ATI x     - info x[3=charge, 2=ttyinfo, 0=1=nothing]
      ATO       - go online (result 3 if fail)
      ATQ       - responses x[0=on, 1=off]
      ATS x     - set register x
                  params:
                      .b=n  set bit state b=0..7, n=0..1
                      .b?   inquire bit b=0..7
                      =n    set state n
                      ?     inquire state
                  regs:
                      0=ringata
                      1=ringcnt
                      2=esc       ('+')
                      3=cr        (13)
                      4=lf        (10)
                      5=bs        (8)
                      7=waitc
                      12=misc, bits:
                           0=resp
                           1=respnum
                           2=echo
                           3=dcd
                           4=cts
                           5=dtrr
                           6=dsr
                           7=cppp
                      13=misc2, bits
                           1=t70
                           2=dtrhup
                           3=respxt
                           4=cidonce
                           5=t70_ext
                           6=rung
                           7=display
                      14=l2prot
                      15=l3prot
                      16=psize
                      17=wsize
                      18=si1
                      19=si2
                      20=si1i
                      21=plan
                      22=screen
                      23=misc3, bits
                           0=cpn
                           1=cpnfcon
      ATV       - responses x[0=numeric, 1=text]
      ATZ       - hook on and reset regs
      AT+FCLASS - params:
                      ?     inquire state
                      =n    set state n
                              0: l2prot=X71I
                                 l3prot=TRANS
                                 si1=4 2: l2prot=FAX
                                 xmitsize=psize*16 
                              2: l3prot=FAX
                                 si1=1
                                 xmitsize=psize*16
                              8: l2prot=X75I
                                 l3prot=TRANS
                                 si1=5
                                 xmitsize=VBUF (0x3E0)
                              ?: outputs '0,2,8' (possible states)
      AT+V      - params (audio only):
                   si1&1 must be 1 (at+fclass=8 or 2).                   
                      NH?    outputs '1' (auto hangup feature)
                      NH=1   does nothing
                      NH=?   outputs '1'
                      IP     reset all voice parameters
                              vpar[0] = 2 (phone line, accept incoming)
                              vpar[1] = 0 (silence detection level = none)
                              vpar[2] =70 (silence interval = 7 seconds)
                              vpar[3] = 2 (compression type = adpcm-2)
                              vpar[4] = 0 (dtmf detection level = softcode)
                              vpar[5] = 8 (dtmf interval = 8 * 5ms)
                      LS?    outputs contents of vpar[0]
                      LS=?   outputs '0,2' (possiblestates)
                      SD     silence detection
                      SD?    outputs vpar[1],vpar[2]
                      SD=a,b sets vpar[1]=a, vpar[2]=b   (silence detection)
                      SD=?   displays '<0-31>,<0-255>' (possible states)
                      SM?    displays <vpar[3]>,<vpar[1]><8000>    (compression)
                      SM=n   sets state
                               2=ADPCM;2;0;(8000)
                               3=ADPCM;3;0;(8000)
                               4=ADPCM;4;0;(8000)
                               5=ALAW;8;0;(8000)
                               6=ULAW;8;0;(8000)
                      SM=?   displays possible states (the above list)
                      LS=n   sets vpar[0] = n
                               0=select device 0
                               2=accept incoming calls
                      RX     start recording
                             vpar[0] must be 2.
                             gives result 8 of not online
                             gives result 1 of ok.
                      TX     start sending
                             vpar[0] must be 2.
                             gives result 8 of not online
                             gives result 1 of ok.
                             (sending and recording CAN be
                              simultaneous. (AT+VRX+VTX))
                      DD?    displays <vpar[4]>,<vpar[5]>
                      DD=a,b ~sets vpar[4]=a, vpar[5]=b    (DTMF detection)
                             must be online when doing this.
                      DD=?   displays '<0-15>,<0-255>' (possible states)
      
      AT+S      - suspend (takes msnstr) ?
      AT+R      - resume (takes msnstr) ?
      AT+M      - message to modem ?
      AT&B n    - sets buffer size to n (psize)
      AT&D n    - sets low-dcd behaviour
                   0=clear BIT_DTRHUP, clear BIT_DTRR
                   2=set BIT_DTRHUP, clear BIT_DTRR
                   3=set BIT_DTRHUP, set BIT_DTRR
      AT&E ...  - set eaz/msn
      AT&F      - does ATZ if not online. If online, does error.
      AT&L ...  - set numbers to listen on
                  accepted chars:
                     0-9
                     ,-*[]?;
      AT&R n    - set V1.110 bitrate adaption:
                    0:    set l2prot = X75I
                              xmitsize = psize*16
                              si2 = 0
                    9600: set l2prot = V2_V11096
                              xmitsize = psize*16/10
                              si2 = 197
                    19200: set l2prot = V2_V11019
                              xmitsize = psize*16/10
                              si2 = 199
                    38400: set l2prot = V2_V11038
                              xmitsize = psize*16/10
                              si2 = 198 (no existing standard)
                  sets T.70 off
                  sets service 7 (si1 |= 4)
      AT&S n    - set window size, n=1..9   (??)
      AT&V      - displays registers (S00..S23, EAZ/MSN, listen-numbers)
      AT&W0     - writes profile
      AT&X n    - switch to btx-mode and T.70
                    0: set xmitsize = psize*16
                           t70 = not t70, not ext                    
                    1: set l2prot = X75I
                           si2 = 0
                           si1 = 4
                           xmitsize = 112
                           t70 = t70, not ext
                    2: set l2prot = X75I
                           si2 = 0
                           si1 = 4
                           xmitsize = 112
                           t70 = t70, ext
*/
	/* Send abort transmission code */
	Send(DLE);
	Send(0x03);
	/* Send abort receiving code also */
	Send(DLE);
	Send(DC4);

	for(;;)
	{
		fd_set fds;
	
		struct timeval tv = {0,250000};
		
		FD_ZERO(&fds);
		FD_SET(fd, &fds);
		FD_SET(0,  &fds);
		
		if(select(fd+1, &fds,NULL,NULL, &tv) > 0)
		{
			char c;
			if(FD_ISSET(fd, &fds))
			{
				char Buf[512];
				int count;
				while((count = read(fd, Buf, 512)) > 0)
					for(char *s=Buf; count; count--,s++)
					{
						c = *s;
						if(c==0x10)write(1, "<DLE>", 5);
						else if(c==0x13)write(1, "<ETX>", 5);
						else if(c==0x14)write(1, "<DC4>", 5);
						else if(c<32 && c!='\n')
						{
							char Buf[32];
							sprintf(Buf, "\33[1m<%02X>\33[m", (unsigned char)c);
							write(1, Buf, strlen(Buf));
						}
						else write(1, &c, 1);
					}
			}
			if(FD_ISSET(0, &fds))
				while(read(0, &c, 1) == 1)
				{
					if(c==27)return;
					if(c==127)c=8;
					write(fd, &c, 1);
				}
		}
	}	
}

static void InitAnswerer(const char *mynumber)
{
	Send(
		"AT+FCLASS=8"  //enter phone(voice) mode.
		  " +VIP"      //reset voice params.
		  " S0=0"      //don't answer automatically.
		  " E0"        //don't emulate me, please.
		  " H0"        //start offline.
		  " Q0"        //responses please.
		  " V1\n"      //responses in text please.
		               //(numeric mode does not offer caller number).
	    "AT&E");    Send(mynumber);  // Set MSN/EAZ.
	Send("\nAT&L"); Send(mynumber);  // Set listening number.
	Send("\n"
	    "AT+VSM=6\n");  //enter 8-bit ULAW-mode (8 kHz)
}

static void GetTextReply(char *buf, int maxlen)
{
	char *s;
	fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) & ~O_NONBLOCK);
	for(;;)
	{
		char c;
		int left;
		for(left=maxlen, s=buf; left && read(fd, &c, 1)==1 && c!='\n'; left--)
		{
			if(c=='\r')continue;
			*s++ = c;
		}
		if(s!=buf)break;
	}
	*s=0;
	printf("Got '%s'\n", buf);
	fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
}

static const char *TransCaller(const char *number)
{
	/* Returns the caller name. */
	
	return number;
}

static char **lines;
static int linecount;

static void InitCode(void)
{
	char *s;
	
	/* get num of lines */
	for(linecount=0, s=code; *s; s++)if(*s=='\n')linecount++;
	lines = (char **)malloc(linecount * (sizeof(char *)));
	for(linecount=0, s=code; *s; *s++ = 0)
	{
		if(*s != ';')lines[linecount++] = s;
		while(*s!='\n' && *s)s++;
		if(!*s)break;
	}
}
static int FindCodeLabel(const char *lab)
{
	int l;
	for(l=0; l<linecount; l++)
	{
		const char *s = lines[l];
		if(*s==':' && !strcmp(s+1, lab))break;
	}
	return l;
}

/* BLOCKING! */
static void CheckInput(int addifno = -1)
{
	static bool dle = false;
	int maxfd = fd;
	
	fd_set fds;
		
	FD_ZERO(&fds);
	FD_SET(fd, &fds);
	
	if(addifno >= 0)
	{
		FD_SET(addifno, &fds);
		if(addifno > maxfd)maxfd = addifno;
	}
		
	select(maxfd+1, &fds,NULL,NULL, NULL);
	
	if(FD_ISSET(fd, &fds))
	{
		char Buf[512];
		int count;
		while((count = read(fd, Buf, 512)) > 0)
		{
//			printf("Got %u bytes: '", count);
			for(char *s=Buf; count; count--,s++)
			{
				unsigned char c = *s;
				
				if(dle)
				{
        			/* TULKITSE VASTAANOTETTU ESCAPE */
        			
        		//	if(c >= '0' && c <= '9')
        		//	{
        				printf("\33[1m<Got DTMF %c>\33[m\n", c);
        		//	}
        			dle = false;
        		}
        		else
        		{
        			if(c == DLE)
        				dle = true;
        		}
			}
//			printf("'\n");
		}
	}
}

static void PipeCommand(const char *cmd)
{
	FILE *fp;
	printf("Popening '%s'\n", cmd);
	fp = popen(cmd, "r");
	if(!fp)
	{
		printf("popen failed\n");
		return;
	}
	unsigned count=0;
	/* Send abort receiving and abort transmission */
	/* Start sending */
	for(;;)
	{
		CheckInput(fileno(fp));
		
		int c = fgetc(fp);
		if(c < 0)break;
		
		++count;
		
		if(c==DLE)
		{
			Send(c);
			/* DLE must be escaped. */
		}
		Send(c);
	}
	printf("Pipe closed, %u bytes of audio transmitted\n", count);
	pclose(fp);
}

/*
     send:
      DLE-DLE = output DLE                      ^P^P
      DLE-ETX = end of data, vonline |= 4       ^P^S
      DLE-DC4 = abort RX,    vonline &= ~1      ^P^T
      DLE-0x03= abort TX                        ^P^C
      DLE-'s' and
      DLE-'q' = silence (?)
      
     received:
      DLE-num = received DTMF code (0..15) ?
*/      

static void RunCode(void)
{
/*
	: label
	p line
	P runcommand
	# getnum
	Q
	I
	\n label
*/
	int line = FindCodeLabel("INIT");
	const char *prog;
	
	/* Start receiving and sending */
	Send("AT+VRX +VTX\n"); 
	
	for(;;)
	{
		const char *s = lines[line];
		printf("Hmm, translating '%s'\n", s);
		switch(*s)
		{
			case 'p':
				prog = s+2;
				break;
			case 'Q':
				goto breakbreak;
			case 'P':
			{
				char *q = (char *)malloc(strlen(s) + strlen(prog) + 10);
				sprintf(q, "%s %s", prog, s+2);
				PipeCommand(q);
				free(q);
				break;
			}
		}
		++line;
	}
breakbreak:
	for(;;)
		CheckInput();

	/* Send abort transmission code */
	Send(DLE);
	Send(0x03);
	/* Send abort receiving code also */
	Send(DLE);
	Send(DC4);
}

static void Answerloop(void)
{
	char Buf[256];
	char caller[256];
	
	struct
	{	
		char caller[256];
		int ringcount;
	} callstatus;
	
	callstatus.ringcount = 0;
	
	for(;;)
	{
		GetTextReply(Buf, (sizeof Buf)-1);
		if(strstr(Buf, "CALLER NUMBER"))
		{
			char caller[256];
			strncpy(caller, strchr(Buf, ':')+2, sizeof(caller)-1);
			caller[sizeof(caller)-1]=0;
		}
		if(!strcmp(Buf, "RING"))
		{
			int num;
			Send("ATS1?\n");
			GetTextReply(Buf, (sizeof Buf)-1);
			num = atoi(Buf);
			
			if(num <= callstatus.ringcount)
			{
				/* Write this call up */
			}

			strcpy(callstatus.caller, TransCaller(caller));
			
			if(num == NumRingsToWait)
				Send("ATA\n");
		}
		if(!strcmp(Buf, "VCON"))
		{
			RunCode();
		}
	}
}

int main(void)
{
	Open(ISDNDevice);
	
	InitTerm(ownnumber);
	Term();
	
	InitCode();
	InitAnswerer(ownnumber);
	Answerloop();
	
	Close();
	
	return 0;
}
